@@ -16,6 +16,10 @@ struct hs20_svc {
 	FILE *debug_log;
 	sqlite3 *db;
 	const char *addr;
+	const char *test;
+	const char *imsi;
+	const char *eap_method;
+	const char *id_hash;
 };
 
 
